
/*--------- threads support ------------------------- g.f. -----*/
/*--------- lower half of the Oberon Threads module             */
/*
 * The PowerPC Oberon compiler follows the old MacOS ABI which
 * is different from the ABI used in Linux and MacOS X. Because of
 * this the interfacing between C and Oberon is a little bit
 * tricky and this module has to be compiled with
 *	-DPowerPC_Oberon_Compiler
 */


#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <signal.h>
#include "Threads.h"


extern int suid_root;

static _thr_ mainthread = 0;

static int prio_high, prio_low;

static pthread_mutex_t at_mutex;
static pthread_mutex_t prio_mutex;

static struct sched_param oldparam;
static int oldpolicy;

#ifdef USE_THREADS

#ifdef PowerPC_Oberon_Compiler
	extern void CallOberon( int pc, int sb );
#endif



_mut_
_mtx_init( ) {
	
	_mut_ mtx;

	mtx = (_mut_)malloc( sizeof(pthread_mutex_t) );
	pthread_mutex_init( mtx, NULL );
	return mtx;
}



void _mtx_destroy(_mut_ mtx) {
	
	(void)pthread_mutex_destroy( mtx );
	free( mtx );
}



void _mtx_lock(_mut_ mtx) {
	
	(void)pthread_mutex_lock( mtx );
}



void _mtx_unlock(_mut_ mtx) {
	
	(void)pthread_mutex_unlock( mtx );
}



static void suspend_handler(int sig) {

	sigpause( SIGUSR2 );
}



static  void resume_handler(int sig) {
}



static void *starter(void *p) {
	_thr_ me = pthread_self();
	oberon_proc proc = (oberon_proc)p;
	sigset_t old, new;
	struct sched_param param;

	sigfillset( &new );
	sigdelset( &new, SIGILL );
	sigdelset( &new, SIGTRAP );
	sigdelset( &new, SIGFPE );
	sigdelset( &new, SIGBUS );
	sigdelset( &new, SIGSEGV );
	sigdelset( &new, SIGTERM );
	sigdelset( &new, SIGUSR1 );
	sigdelset( &new, SIGUSR2 );
	pthread_sigmask( SIG_SETMASK, &new, &old );

	signal( SIGUSR1, suspend_handler );
	signal( SIGUSR2, resume_handler );

	if ( suid_root != 0) {
	    param.sched_priority = 0;
	    pthread_setschedparam( me, SCHED_OTHER, &param );
	}

#ifdef PowerPC_Oberon_Compiler
	CallOberon( proc->pc, proc->sb );
#else
	proc();
#endif
	return NULL;
}



#ifdef PowerPC_Oberon_Compiler
_thr_ 
_thr_start( int PC, int SB, int len ) {
	
	oberon_proc p = (oberon_proc)malloc(8);
	_thr_ id;
	pthread_attr_t attr;
	p->pc = PC; p->sb = SB;
#else
_thr_ 
_thr_start( oberon_proc p, int len ) {
	
	_thr_ id;
	pthread_attr_t attr;
#endif
	 
	if (len < PTHREAD_STACK_MIN) len = PTHREAD_STACK_MIN;
	pthread_attr_init( &attr );
	if ( suid_root ) {
	    pthread_attr_setstacksize( &attr, len );
	}
	if (pthread_create( &id, &attr, starter, p ) != 0) return 0;
	return id;
}



_thr_ _thr_this() {

	return pthread_self();
}



void _thr_sleep(int ms) {

	struct timespec sltime, rem;

	sltime.tv_sec = ms/1000;
	sltime.tv_nsec = 1000000*(ms%1000);
	nanosleep( &sltime, &rem );
}



void _thr_pass( ) {

	_thr_sleep( 10 );
}



void _thr_exit( ) {

	pthread_exit( 0 );
}



void _thr_sendsig( _thr_ thr, int sig ) {

	pthread_kill( thr, sig );
}



void _thr_suspend(_thr_ thr) {

	pthread_kill( thr, SIGUSR1 );
}



void _thr_resume(_thr_ thr) {

	pthread_kill( thr, SIGUSR2 ); 
}



void _thr_setprio(_thr_ thr, int prio) {

	struct sched_param param;
	int policy;

	if ( suid_root == 0) return;

	pthread_mutex_lock( &prio_mutex );
	seteuid(0);
	pthread_getschedparam( thr, &policy, &param );
	param.sched_priority = prio;
	if (pthread_setschedparam( thr, SCHED_FIFO, &param ) != 0)
		perror("pthread_setschedparam");
	seteuid( getuid());
	pthread_mutex_unlock( &prio_mutex );
}



int _thr_getprio(_thr_ thr) {

	struct sched_param param;
	int policy;

	if ( suid_root == 0) return 0;

	pthread_mutex_lock( &prio_mutex );
	pthread_getschedparam( thr, &policy, &param );
	pthread_mutex_unlock( &prio_mutex );
	return ( param.sched_priority );
}



void _thr_maxprio( ) {
	
	struct sched_param newparam;
	_thr_ this;

	if ( suid_root == 0) return;

	pthread_mutex_lock( &at_mutex );
	seteuid(0);
	this = pthread_self();
	pthread_getschedparam( this, &oldpolicy, &oldparam );
	newparam.sched_priority = prio_high;
	if (pthread_setschedparam( this, SCHED_FIFO, &newparam ) != 0)
		perror("pthread_setschedparam");
	seteuid( getuid());
}



void _thr_oldprio( ) {
	
	struct sched_param param;
	int policy;

	if ( suid_root == 0) return;

	seteuid(0);
	param = oldparam;
	policy = oldpolicy;
	pthread_mutex_unlock( &at_mutex );
	if (pthread_setschedparam( pthread_self(), policy, &param ) != 0)
		perror("pthread_setschedparam");
	seteuid( getuid());
}



void _thr_kill(_thr_ thr) {

	if (thr != mainthread) {
		pthread_detach( thr );
		if (thr == pthread_self())
			pthread_exit( 0 );
		else
			pthread_cancel( thr );
			pthread_kill( thr, SIGCONT );
	}
}


/* _thr_initialize returns 0 (FALSE) if no threads support is compiled in */

int _thr_initialize( int *low, int* high ) {
	
	mainthread = pthread_self();
	if ( suid_root == 0) {
		prio_high = sched_get_priority_max(SCHED_RR);
		prio_low = sched_get_priority_min(SCHED_RR);
		*low = prio_low;
		*high = prio_high;
	    	/*param.sched_priority = prio_high;*/
	    	pthread_setschedparam( mainthread, SCHED_RR, NULL );
	}else{
		prio_high = sched_get_priority_max(SCHED_FIFO);
		prio_low = sched_get_priority_min(SCHED_FIFO);
		*low = prio_low;
		*high = prio_high;	

		pthread_mutex_init( &at_mutex, NULL );
		pthread_mutex_init( &prio_mutex, NULL );
	}
	
	return 1;
}
#else
_mut_	_mtx_init( ) {return NULL;}
void	_mtx_destroy( _mut_ mtx ) { }
void	_mtx_lock(    _mut_ mtx ) { }
void	_mtx_unlock(  _mut_ mtx ) { }

#ifdef PowerPC_Oberon_Compiler
_thr_	_thr_start( int PC, int SB, int len ) {return NULL;}
#else
_thr_	_thr_start( oberon_proc p, int len ) {return 0;}
#endif

_thr_	_thr_this( ) {return 0;}
void	_thr_sleep( int ms ) { }
void	_thr_pass( ) { }
void	_thr_exit( ) { }
void	_thr_sendsig( _thr_ thr, int sig ) { }
void	_thr_suspend( _thr_ thr ) { }
void	_thr_resume(  _thr_ thr ) { }
void	_thr_setprio( _thr_ thr, int prio ) { }
int 	_thr_getprio( _thr_ thr ) {return 0;}
void	_thr_maxprio( ) { }
void	_thr_oldprio( ) { }
void	_thr_kill( _thr_ thr ) { }

int	_thr_initialize( int *low, int* high ) {

	printf("Oberon boot loader has been compiled without Threads support!\n");
	return 0;
}
#endif


