
/*--------- threads support ------------------------- g.f. -----*/
/*--------- lower half of the Oberon Threads module             */

#include <stdio.h>
#include <limits.h>
#include "Threads.h"

static int suid_root;
static _thr_ mainthread = 0;

#ifdef USE_THREADS
_mut_ _mtx_init( ) {
    _mut_ mtx;

    mtx = (_mut_)malloc( sizeof(mutex_t) );
    mutex_init( mtx, USYNC_THREAD, NULL );
    return mtx;
}

void _mtx_destroy(_mut_ mtx) {
    (void)mutex_destroy( mtx );
    free( mtx );
}


void _mtx_lock(_mut_ mtx) {
    (void)mutex_lock( mtx );
}


void _mtx_unlock(_mut_ mtx) {
    (void)mutex_unlock( mtx );
}


_thr_ _thr_start( oberon_proc p, int len ) {
    _thr_ id;
    sigset_t orig, new;
    int	err;

    if ((len != 0) && (len < 16*1024)) {
        len =  16*1024;
    }

    sigfillset( &new );
    sigdelset( &new, SIGILL );
    sigdelset( &new, SIGTRAP );
    sigdelset( &new, SIGFPE );
    sigdelset( &new, SIGBUS );
    sigdelset( &new, SIGSEGV );
    sigdelset( &new, SIGTERM );
    thr_sigsetmask( SIG_SETMASK, &new, &orig );
    err = thr_create( NULL, len, p, NULL, THR_BOUND | THR_NEW_LWP, &id );
    thr_sigsetmask( SIG_SETMASK, &orig, NULL );
    if (err != 0)
        return 0;
    return id;
}


_thr_ _thr_this() {
    return thr_self();
}


void _thr_sleep(int ms) {
    struct timespec sltime, rem;

    sltime.tv_sec = ms/1000;
    sltime.tv_nsec = 1000000*(ms%1000);
    nanosleep( &sltime, &rem );
}


void _thr_pass( ) {
    _thr_sleep( 1 );
}

void _thr_exit() {
    thr_exit( 0 );
}


void _thr_suspend(_thr_ thr) {
    thr_suspend( thr );
}


void _thr_resume(_thr_ thr) {
    thr_continue( thr );
}


void _thr_setprio(_thr_ thr, int prio) {
    thr_setprio( thr, prio );
}


int _thr_getprio(_thr_ thr) {
    int prio;

    thr_getprio( thr, &prio );
    return ( prio );
}


void _thr_maxprio( ) { }


void _thr_oldprio( ) { }


void _thr_kill(_thr_ thr) {
    if (thr != mainthread) {
        if (thr == thr_self())
            thr_exit( 0 );
        else 
            thr_kill( thr, SIGTERM );
    }
}


/* thr_initialize returns 0 (FALSE) if the program has
   been compiled without threads suport. If the program
   has no suid root privilleges, priorities are disabled
   and low and high both return 0. */

int _thr_initialize( int *low, int* high ) {
    mainthread = thr_self();
    *low = 0;  *high = 100;
    return 1;
}

#else

_mut_	_mtx_init( ) { }
void	_mtx_destroy( _mut_ mtx ) { }
void	_mtx_lock(    _mut_ mtx ) { }
void	_mtx_unlock(  _mut_ mtx ) { }

#ifdef PowerPC_Oberon_Compiler
_thr_	_thr_start( int PC, int SB, int len ) { }
#else
_thr_	_thr_start( oberon_proc p, int len ) { }
#endif

_thr_	_thr_this( ) { }
void	_thr_sleep( int ms ) { }
void	_thr_pass( ) { }
void	_thr_exit( ) { }
void	_thr_sendsig( _thr_ thr, int sig ) { }
void	_thr_suspend( _thr_ thr ) { }
void	_thr_resume(  _thr_ thr ) { }
void	_thr_setprio( _thr_ thr, int prio ) { }
int 	_thr_getprio( _thr_ thr ) { }
void	_thr_maxprio( ) { }
void	_thr_oldprio( ) { }
void	_thr_kill( _thr_ thr ) { }

int	_thr_initialize( int *low, int* high ) {

	printf("Oberon boot loader has been compiled without Threads support!\n");
	return 0;
}
#endif


