
/***********
 * Modules0.h
 * C interface of the Oberon module loader for the IBM RISC System/6000
 * Marc-Michael Brandis, 26.4.91 / 31.3.92
 *
 * G. Feldmann 20.1.97,	term handler added <<
 *	       20.2.98, new Module structure
 *	       20.10.98 new LinkEntry structure
 *
 ***********/

#ifndef _I_MODULES
#define _I_MODULES

#include "OberonTypes.h"

/* values for res of ThisMod */
#define done 0
#define fileNotFound 1
#define invalidObjFile 2
#define invalidKey 3
#define notEnoughMemory 4
#define cmdNotFound  5
#define modNotFound  6
#define cyclicImport  7
#define modInUse 10
#define libNotFound 11
#define symNotFound 15

/* sizes */
#define ModuleNameLength 32
#define CommandNameLength 24
#define MessageLength 256

/* types */

typedef PROC Command;

typedef CHAR ModuleName[ModuleNameLength];
typedef CHAR CommandName[CommandNameLength];
typedef CHAR Message[MessageLength];


typedef struct command_descriptor {
	CommandName name;
	INTEGER	offset;
	INTEGER	dummy;		/* force C to align8!  << g.f. */
} CommandDesc;

typedef struct trap_desc {
	INTEGER	offset;
	INTEGER	trapno;
} TrapDesc;

typedef struct export_desc {
	LONGINT	fp;
	LONGINT	adr;
	INTEGER	nofExp;
	LONGINT	dsc;	/* ExportPtr */
} ExportDesc;

typedef ExportDesc* ExportPtr;

typedef struct module_desc {
	struct module_desc*	link;
	ModuleName	name;
	LONGINT		key;
	INTEGER		refCnt;
	LONGINT		dataSize;
	LONGINT		blockSize;
	LONGINT		refSize;
	INTEGER 	conSize;
	INTEGER		codeSize;
	INTEGER		nofEntries;
	INTEGER		nofCmds;
	INTEGER		nofImps;
	INTEGER		nofTds;
	INTEGER		nofPtrs;
	INTEGER		nofTraps;
	ADDRESS		block;
	LONGINT		SB;
	LONGINT		PC;
	LONGINT		entries;		/* POINTER TO ARRAY OF INTEGER */
	LONGINT		commands;		/* POINTER TO ARRAY OF CommandDesc */
	LONGINT		imports;		/* POINTER TO ARRAY OF Module */
	LONGINT		typedescs;		/* POINTER TO ARRAY OF ADDRESS */
	LONGINT		pointers;		/* POINTER TO ARRAY OF LONGINT */
	LONGINT		traps;			/* POINTER TO ARRAY OF TrapDesc */
	LONGINT		refs;			/* POINTER TO ARRAY OF CHAR */
	LONGINT		syms;
	LONGINT		publics;
	LONGINT		privates;
	INTEGER		nofimp;
	INTEGER		_dum1;			/* force align4 for next member */
	ADDRESS		import;
	INTEGER		nofstrc;
	INTEGER		_dum2;			/* force align4 for next member */
	ADDRESS		structs;
	INTEGER		nofreimp;
	INTEGER		_dum3;			/* force align4 for next member */
	ADDRESS		reimp;
	ExportDesc	export;
	PROC		term;			
} ModuleDesc;

typedef ModuleDesc* Module;

/* if NULL is not defined, we do it here */
#ifndef NULL
#define NULL ((void*)0L)
#endif

#define NIL ((void*)0L)


typedef Tag (* CreateTDproc)(Module, CHAR*, LONGINT, LONGINT*, LONGINT, Tag, LONGINT, PROC*, LONGINT);

extern void ModulesInit(void);
extern void LoadKernel(void);

#endif
