/*
 *  C interface for X11 callbacks to Oberon procedures
 *		and X11 procedures with more than 8 parameters
 *  G. Feldmann 27.1.97
 */
#include <stdio.h>
#include <X11/Xlib.h>

typedef struct _v2 {
	uint a, b;
} *v2;


typedef struct _v3 {
	uint a, b, c;
} *v3;

typedef struct _v4 {
	uint a, b, c, d;
} *v4;


extern void* oberonXErrorHandler;
extern void* oberonXIOErrorHandler;

static int X11ErrorHandler( Display *d, XErrorEvent *err ) {
	long *p,pc,sb;

	p = (long *)((long)oberonXErrorHandler);
	pc = *p++; sb = *p;
	CallOberon(pc, sb, d, err );
}

static int X11IOErrorHandler( Display *d ) {
	long *p,pc,sb;

	p = (long *)((long)oberonXIOErrorHandler);
	pc = *p++; sb = *p;
	CallOberon(pc, sb, d );	
}


void SetupXErrorHandlers() {
	
	XSetErrorHandler(X11ErrorHandler);
	XSetIOErrorHandler(X11IOErrorHandler);	
}


Window gXCSW(Display *d, Window pw, v2 pos, v2 size, ulong bw, ulong b, ulong bg ) {
	return(XCreateSimpleWindow(d,pw,pos->a,pos->b,size->a,size->b,bw,b,bg));
}

void gXCA(Display *d, Drawable src, Drawable dest, GC gc, v2 sp, v2 size, v2 dp) {
	XCopyArea(d,src,dest,gc,sp->a,sp->b,size->a,size->b,dp->a,dp->b);
}

void gXDA(Display *d, Drawable w, GC gc, v2 pos, v2 size, int a1, int a2) {
	XDrawArc(d,w,gc,pos->a,pos->b,size->a,size->b,a1,a2);
}

void gXFA(Display *d, Drawable w, GC gc, v2 pos, v2 size, int a1, int a2) {
	XFillArc(d,w,gc,pos->a,pos->b,size->a,size->b,a1,a2);
}


void gXCP(Display *d, Drawable src, Drawable dest, GC gc, v2 sp, v2 size, v2 dp, ulong plane) {
	XCopyPlane(d,src,dest,gc,sp->a,sp->b,size->a,size->b,dp->a,dp->b,plane);
}


void gXGG(Display *d, Drawable src, ulong *root, v2 pos, v2 size, uint *ow, uint *depth) {
	XGetGeometry(d,src,root,(uint*)pos->a,(uint*)pos->b,
		     (uint*)size->a,(uint*)size->b,ow, depth);
}

void gXQP(Display *d, Window w, v4 b1, v3 b2) {
	XQueryPointer(d,w,(Window*)b1->a,(Window*)b1->b,(uint*)b1->c,(uint*)b1->d,
		      (int*)b2->a,(int*)b2->b,(uint*)b2->c);
}

XImage *gXCI(Display *d, Visual *v, v4 dfod, v2 size, ulong bP, ulong bpL) {
	return XCreateImage(d,v,dfod->a,dfod->b,dfod->c,(char*)dfod->d,
			    size->a,size->b,bP,bpL);
}

XImage *gXGSI(Display *d, Drawable w, v4 src, v2 pmf, XImage *dim, v2 dst) {
	return XGetSubImage(d,w,src->a,src->b,src->c,src->d,pmf->a,pmf->b,
			    dim,dst->a,dst->b);
}

void gXPI(Display *d, Drawable w, GC gc, XImage *image, v2 src, v2 dst, v2 size) {
	XPutImage(d,w,gc,image,src->a,src->b,dst->a,dst->b,size->a,size->b);
}

int gXSMPI(Display *d, Drawable w, GC gc, XImage *image, v4 sd, v2 size, int ev) {
	return XShmPutImage(d,w,gc,image,sd->a,sd->b,sd->c,sd->d,size->a,size->b,ev);
}

void gXGWP(Display *d, Window w, v4 pold, v2 rtt, v4 fnbp) {
	XGetWindowProperty(d,w,pold->a,pold->b,pold->c,pold->d,
			   rtt->a,(Atom*)rtt->b,(int*)fnbp->a,
			   (ulong*)fnbp->b,(ulong*)fnbp->c,
			   (unsigned char**)fnbp->d);
}

void gXWP(Display *d, Window sw, Window dw, v4 src, v2 dst) {
	XWarpPointer(d,sw,dw,
		     src->a,src->b,src->c,src->d,
		     dst->a,dst->b);
}
