/*
    ETH Oberon (Solaris SPARC) runtime loader
    Copyright (c) J. Templ, 1993
    
    2000.03.19	G. Feldmann	release 2.3.6
    2000.04.28	G. Feldmann	support for Kernel.NewArr added
    2000.07.05	G. Feldmann	new object format
    2000.08.29	"	switched threads support from
    			posix threads to native threads
    2001.02.20	" 	fixed a bug in ReadTypes
    2001.05.21	"	new obj flag
    2001.06.08	"	support for HUGEINT added (Oberon_ModH, ...)

    without theads:
      gcc -DSOARIS oberon.c -o sparc.oberon -ldl -lm 
    
    with theads:
      gcc -DSOARIS -DUSE_THREADS oberon.c -o sparc.oberon -ldl -lrt -lm 
*/

#define _REENTRANT	1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <dlfcn.h>
#include <malloc.h>
#include <sys/ieeefp.h>
#include <floatingpoint.h>
#include <math.h>
#include <setjmp.h>
#include <signal.h>
#include <thread.h>
#include <limits.h>
#include <sys/ucontext.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/mman.h>

#include "Threads.h"


/* Oberon_ThisMod result codes */
#define done 0
#define fileNotFound 1
#define invalidObjFile 2
#define corruptedObjFile 4
#define notEnoughSpace 7
#define cyclicImport 9
#define incompImport 16


/* Oberon_ThisCommand result codes */
#define cmdNotFound 5

/* Oberon_Free result codes */
#define moduleNotFound 6
#define refCntNotZero 8



#define EUEnd 0
#define EURecord 1
#define EUobjScope 0
#define EUrecScope 1
#define EUerrScope -1

#define EUProcFlag 0x80000000
#define EUTypFlag  0xC0000000


#define	false	0
#define	true	1

/* max record extension level */
#define maxlevel 7

#define MaxRecs	64


extern float __floatdisf();
extern float __floatdidf();
extern long long __dtoll();
extern long long __ftoll();

typedef struct {
    u_char m, n; 
    long   fixups;
} LinkDesc; 
 
struct td;
typedef struct tsd {
    struct tsd* next;
    struct td*  tag;
} TypesDesc, *Types;

typedef struct {
    char name[32];
    long entry;
}CommandDesc, *Command;

typedef struct ed {
    long       fp;
    long       adr;
    long       nofExp;
    struct ed  *dsc;
} ExportDesc, *Export;

typedef struct md {
    struct md  *next;
    char       name[32];
    char       init;
    long       refcnt;
    long       dataSize, conSize, codeSize, refSize;
    long       nofEntries, nofCmds, nofImps, nofTds, nofPtrs;
    long       *entries;
    Command    cmds;
    long       *ptrTab;
    Types      types;
    struct md  **imports;
    char       *data, *code, *refs;
    Export     export;
    char       *term;
} ModuleDesc, *Module;

typedef struct td {
    long       blksize, recsize, reserved;
    Module     module;
    char       name[32];
    short      nofmthds, level;
    long       base[maxlevel + 1];
    long       ptr[1];
} TypeDesc, *Type;

typedef struct {
    char name[32]; 
    long key;
} impMod; 

typedef struct {
    FILE  *handle;
    char  name[256];
    int   pos;
} FileDesc, *File;

int suid_root=0;

/* command line options */
static long ac;
static char **av;

static long pagesize;

/* 9 of 14 SPARC-Oberon runtime routines */
static long mul, mulH, newrec, newsys, newarr, RtoH, LRtoH, HtoR, HtoLR;

/* the search path for loading Oberon object files */
static char *OBERON;

/* module list anchor */
static Module modules;

/* for loader error messages */
static char n1[64], n2[64], n3[64];

static void Oberon_ThisMod(char*, long, Module*, short*);
static void Oberon_ThisCommand(Module, char*, long, long*, short*);
static void Oberon_Free(char*, long, char, short*);


static void* Oberon_dlopen(char *libname, long length, long mode) { 

    return (dlopen(libname, 1));
}

/*------------------ trap handling ------------------------------*/

typedef	void(*traphandler_t)(int, void*, void*);

static traphandler_t	OberonTrapHandler;

static void traphandler(int sig, void *scp, ucontext_t *ucp) {
    
    int i;

    /*printf("traphandler %d %x %x\n", sig, scp, ucp);*/
    OberonTrapHandler(sig, scp, ucp);

    /*printf("returned fron Oberon Trap Handler\n");*/
    return;
}

static void installHandler(int sig) {
    struct sigaction act;
    sigset_t mask;

    sigemptyset(&mask);
    act.sa_mask = mask;
    act.sa_flags =  SA_SIGINFO|SA_ONSTACK|SA_NODEFER;
    act.sa_handler = traphandler;
    if (sigaction( sig, &act, NULL ) != 0) {
        perror("sigaction");
    }
}

static void InstallTrapHandler(traphandler_t p) {
    int i;
    
    /*printf("traphandler = %x\n", p);*/
    OberonTrapHandler = p;

    for (i = 1; i <= 15; i++) {
	if (i != 9) installHandler( i );
    }
}

/*------ dynamic linking -------------------------------------*/

static int errnof() {
    
    return errno;
}


static int o_stat(char* name, void* buf) {
    
    return stat(name, (struct stat *) buf);
}

static int o_lstat(char* name, void* buf) {
    
    return lstat(name, (struct stat *) buf);
}

static int o_fstat(int fd, void* buf) {
    
    return fstat(fd, (struct stat *) buf);
}


static int o_open(char* name, int flags, int mode) {
    int r;

    r = open(name, flags, mode);
    if (r < 0) { 
        fprintf(stderr, "<%s> ", name ); perror("open");  
    }
    return r;
}

static void Oberon_dlsym(void *handle, char *sym, void **adr) { 

    char sym_r[64];

    if      (!strcmp(sym, "n1adr")) *adr=(void*)n1;
    else if (!strcmp(sym, "n2adr")) *adr=(void*)n2;
    else if (!strcmp(sym, "n3adr")) *adr=(void*)n3;
    else if (!strcmp(sym, "thisMod")) *adr=(void*)Oberon_ThisMod;
    else if (!strcmp(sym, "thisCommand")) *adr=(void*)Oberon_ThisCommand;
    else if (!strcmp(sym, "Free")) *adr=(void*)Oberon_Free;
    else if (!strcmp(sym, "modules")) *adr=(void*)modules;
    
/* aux math functions */
    else if (!strcmp(sym, "log")) *adr=(void*)log;
    else if (!strcmp(sym, "exp")) *adr=(void*)exp;

/* install storage manager */
    else if (!strcmp(sym, "setNewRec")) newrec=(long)*adr;
    else if (!strcmp(sym, "setNewSys")) newsys=(long)*adr;
    else if (!strcmp(sym, "setNewArr")) newarr=(long)*adr;

/* install trap handler */
    else if (!strcmp(sym, "InstallTrapHandler")) *adr = (void*)InstallTrapHandler;

/* libc, env.: */
    else if (!strcmp(sym, "argc")) *adr=(void*)ac;
    else if (!strcmp(sym, "argv")) *adr=(void*)av;
    else if (!strcmp(sym, "errno")) *adr=(void*)errnof;
    else if (!strcmp(sym, "dlopen")) *adr=(void*)dlopen;
    else if (!strcmp(sym, "dlclose")) *adr=(void*)dlclose;
    else if (!strcmp(sym, "stat")) *adr=(void*)o_stat;
    else if (!strcmp(sym, "lstat")) *adr=(void*)o_lstat;
    else if (!strcmp(sym, "fstat")) *adr=(void*)o_fstat;

/* threads support */
    else if (!strcmp(sym, "mtxInit")) *adr = (void*)_mtx_init;
    else if (!strcmp(sym, "mtxDestroy")) *adr = (void*)_mtx_destroy;
    else if (!strcmp(sym, "mtxLock")) *adr = (void*)_mtx_lock;
    else if (!strcmp(sym, "mtxUnlock")) *adr = (void*)_mtx_unlock;
    else if (!strcmp(sym, "thrStart")) *adr = (void*)_thr_start;
    else if (!strcmp(sym, "thrThis")) *adr = (void*)_thr_this;
    else if (!strcmp(sym, "thrSleep")) *adr = (void*)_thr_sleep;
    else if (!strcmp(sym, "thrPass")) *adr = (void*)_thr_pass;
    else if (!strcmp(sym, "thrExit")) *adr = (void*)_thr_exit;
    else if (!strcmp(sym, "thrSuspend")) *adr = (void*)_thr_suspend;
    else if (!strcmp(sym, "thrResume")) *adr = (void*)_thr_resume;
    else if (!strcmp(sym, "thrGetPriority")) *adr = (void*)_thr_getprio;
    else if (!strcmp(sym, "thrSetPriority")) *adr = (void*)_thr_setprio;
    else if (!strcmp(sym, "thrMaxPriority")) *adr = (void*)_thr_maxprio;
    else if (!strcmp(sym, "thrOldPriority")) *adr = (void*)_thr_oldprio;
    else if (!strcmp(sym, "thrKill")) *adr = (void*)_thr_kill;
    else if (!strcmp(sym, "thrInitialize")) *adr = (void*)_thr_initialize;
   
    else {
        *adr = dlsym(handle, sym);
    }
    if (*adr == NULL) printf("Symbol not found: %s\n", sym);

}

/*------ module loader -------------------------------------*/

static unsigned char Read1(File f) { 
    unsigned char byte;

    fread(&byte, 1, 1, f->handle);
    f->pos++; 
    return byte; 
}

static short Read2(File f) { 
    short word;
    
    fread(&word, 1, 2, f->handle); 
    f->pos = f->pos + 2;
    return word;
}

static unsigned short Read2u(File f) { 
    unsigned short word;

    fread(&word, 1, 2, f->handle); 
    f->pos = f->pos + 2;
    return word; 
}

static long Read4(File f) { 
    long longword;

    fread(&longword, 1, 4, f->handle); 
    f->pos = f->pos + 4;
    return longword;
}

static void ReadName(File f, char *name) { 
    
    int i = 0; unsigned char c;

    do {
    	c = Read1(f);
	if (c > 0x80) {
		name[i++] = c - 0x80; c = 0;
	}
	name[i++] = c;
    }while (name[i-1] != 0);
}

static long ReadNum( File f ) {
    char c;
    long n, x, s, d;
    
    s = 0; n = 0; 
    fread(&c, 1, 1, f->handle); f->pos++;
    x = c & 0177;
    while (x != c) {
        n = n + (x << s); s = s + 7; 
    fread(&c, 1, 1, f->handle); f->pos++;
        x = c & 0177;
    }
    d = 25 - s;
    if (d >= 0)
        n = n + ((x << 25) >> d);
    else
        n = n + ((x << 25) << -d);
    return n;
}

static long Oberon_Div (u_long x, u_long y) { 

    if ((long) x >= 0) return (x / y);
    else return -((- x - 1 + y) / y); 
}

static long Oberon_Mod (u_long x, u_long y) { 
    u_long m;

    if ((long) x >= 0) return (x % y);
    else { 
        m = (-x) % y;
        if (m != 0) return (y - m); 
        else return 0;
    }
}

typedef unsigned long long	u_llong;
typedef          long long	llong;

static u_llong Oberon_DivH (u_llong x, u_llong y) { 

    if ((llong) x >= 0) return (x / y);
    else return -((- x - 1 + y) / y); 
}


static u_llong Oberon_ModH (u_llong x, u_llong y) { 
    u_llong m;

    if ((llong) x >= 0) return (x % y);
    else { 
        m = (-x) % y;
        if (m != 0) return (y - m); 
        else return 0;
    }
}

static llong Oberon_AshH (llong x, long y) { 
    
    llong e;
    
    if (y < 0) e = x>>(-y); else e = x<<y;
    return e;
}

static int Check(File f, u_char flag, short *res) { 
    unsigned char byte;

    if (*res == done) {
        byte = Read1(f);
        if (byte == flag) { 
	    return true;
        }else{	
            strcpy(n1, f->name);
	    *res = corruptedObjFile;
        }
    }
    return false;
}

static int Check2(File f, u_char flag, u_char flag2, short *res) { 
    unsigned char byte;

    if (*res == done) {
        byte = Read1(f);
        if (byte == flag) { 
	    return true;
        }else{	
	    if (byte == flag2) {
		printf("warnig: %s: old object file format\n", f->name);
		return true;
	    }
            strcpy(n1, f->name);
	    *res = corruptedObjFile;
        }
    }
    return false;
}

/*establish bottleneck interface between C and Oberon */

static void InitGlue(Module mod) { 
    long* adr;

    adr = (long*)((long)mod->data - 4);
    *(adr--) = (long)Oberon_dlsym;
    *(adr--) = (long)Oberon_dlsym;
    *adr = (long)Oberon_dlsym;
}


static void Unlink(Module m) { 
    Module cur, prev;

    if (m == NULL) return;
    prev = NULL;
    if (m == modules) {
        modules = m->next;
    } else {
        cur = modules;
        while (cur != NULL && cur != m) {
	    prev = cur; cur = cur->next;
	}
        if (cur == m) {
            prev->next = m->next;
        }else{
	    prev->next = NULL;
	}
    }
    free( (char*)m->entries );
    free( (char*)((int)m) - 8 );
}

static void FixDataAccess( Module m, long adr, long dataadr ){

    long instr, next, wadr;
    long *code; 
   
    code = (long*)m->code; wadr = adr>>2;
    while (wadr != 0) {
        instr = code[wadr]; next = instr & 0xffff;
    
        /* global data access, one instruction: sethi */
        code[wadr] = (instr & (~0x3fffff)) + dataadr / 1024;
        
        wadr = next>>2;
    }
}

static void FixProc( Module m, long adr, long procadr ){
    
    long instr, dist, next, wadr;
    long *code; 
    
    
    code = (long*)m->code; wadr = adr>>2;
    while (wadr != 0) {
        instr = code[wadr]; next = instr & 0xffff;
        if (instr >= 0x40000000) {
            /* procedure call, one instruction: call */
            dist = procadr - (long)&code[wadr];
            code[wadr] = 0x40000000 | (((u_long)(dist) >> 2) & 0x3fffffff);
        } else {
            /* procedure assignment, 2 instructions: sethi, add */
            code[wadr] = (instr & (~0x3fffff)) + (procadr >> 10);
            instr = code[wadr+1];
            code [wadr+1] = (instr-1023) / 1024 * 1024 + procadr % 1024;
        }
        wadr = next>>2;
    }
}

static void Fixup( Module m, int noflinks, LinkDesc links[1000]) { 

    long i;
    long long RTS[15], procadr;
    long *code;
    unsigned char modnr, entrynr;
    Module im;

    RTS[1]  = mul;
    RTS[2]  = (long)Oberon_Div;
    RTS[3]  = (long)Oberon_Mod;
    RTS[4]  = newrec;
    RTS[5]  = newsys;
    RTS[6]  = newarr;
    RTS[7]  = mulH;
    RTS[8]  = (long)Oberon_DivH;
    RTS[9]  = (long)Oberon_ModH;
    RTS[10] = (long)Oberon_AshH;
    RTS[11] = RtoH;
    RTS[12] = LRtoH;
    RTS[13] = HtoR;
    RTS[14] = HtoLR;

    code = (long *)m->code;
    for (i = 0; i < noflinks; i++) {
        modnr = links[i].m; 
        entrynr = links[i].n; 
	if ((entrynr == 255) & (modnr == 0)) {
            FixDataAccess( m, links[i].fixups, (long)m->data );
        } else {
            if (modnr == 255) { 
                procadr = RTS[entrynr];
		FixProc( m, links[i].fixups, procadr );
            } else { 
		if (modnr == 0) im = m;
		else im = m->imports[modnr-1];
                procadr = (long)(im->code) + im->entries[entrynr];
		FixProc( m, links[i].fixups, procadr );
            }
        }
    }
}


static long TdAdr( Module m, long fp ) {

    Export elem, dsc;
    int i;
    
    dsc = m->export->dsc; elem = dsc;
    i = 0;
    while ((i < m->export->nofExp) & (elem->fp != fp)) {
        i++;
        elem = &dsc[i];
    } 
    if (i < m->export->nofExp)
        return elem->dsc->adr;
    else
        return 0;
}


static Type AllocTD( Module m, int tdadr, int nofmth, int nofptrs ) {
    
    Types t, last;
    long tagadr, gap;
    Type tag;

    t = (Types)calloc(nofmth + maxlevel + nofptrs + 21, 4);
    if (t != NULL) {
        if (m->types == NULL) m->types = t; 
	else {
	    last = m->types;
	    while (last->next != NULL) last = last->next;
	    last->next = t;
	}
        t->next = NULL; last = t;
        tagadr = (long)t + (nofmth+3) * 4; gap = 0;
        while ((tagadr % 16) != 0) {
            tagadr = tagadr + 4; gap = gap + 4;
        }
        *(long*)(m->data + tdadr) = tagadr;
        tag = (Type)tagadr;
        t->tag = tag;
        *((long*)tagadr - 1) = (long)tag + 1;  /* set mark bit ! */
	return tag;
    }else return NULL;
}


static void ReadTypes(File f, Module m, int nofrecs, short *res) { 

    long i, j, k, u, *mthp;
    long recsize, basemod, baseadr, nofptrs;
    long nofmth, nofinhmth, nofnewmth, mthno, entno;
    Type tag, base;
    char name[32];
    Module im;
    int tdadr[MaxRecs];

    *res = done; m->types = NULL;
    for (i = 0; (i < nofrecs) && (*res == done); i++) {
        recsize = Read4(f);
        tdadr[i] = Read2(f);
        basemod = Read2(f); baseadr = Read4(f);
	if (basemod > 0) {
	    /* baseadr contains fp! */
	    im =  m->imports[basemod-1]; baseadr = TdAdr( im, baseadr );
	}
        nofmth = Read2(f);
        nofinhmth = Read2(f); nofnewmth = Read2(f);
        nofptrs = Read2(f);
        ReadName(f, name);

        tag = AllocTD( m, tdadr[i], nofmth, nofptrs );
	if (tag != NULL) {
            tag->blksize = (recsize + 4 + 15) / 16 * 16;
            tag->recsize = recsize;
            tag->reserved = 0;
            strcpy(tag->name, name);
            tag->module = m;
            tag->nofmthds = nofmth;
            for (k = 0; k < nofmth; k++) {
	        *(long*)(((long)tag) - 8 - k*4) = -1;
	    }
            for (k = 0; k < nofnewmth; k++) {
	        mthno = Read2(f); entno = Read2(f);
		*(long*)(((long)tag) - 8 - mthno*4) = entno;
            }
	    for (k = 0; k < nofptrs; k++) 
	    	tag->ptr[k] = Read4(f);
            tag->ptr[k] = ((long)tag) - (long)&tag->ptr[k];
	    
	    /* temp. values, used in fixup */
            tag->base[0] = basemod; 
	    tag->base[1] = baseadr; 
	    tag->base[2] = nofmth;
        }else
            *res = notEnoughSpace;
    }
    
    if (*res != done) return;

    /* now fixup */
    for (i = 0; (i < nofrecs) && (*res == done); /*nothing*/ ) {
        tag = (Type) *(long*)(m->data + tdadr[i]); j = 0;
        basemod = tag->base[0]; baseadr = tag->base[1]; nofmth = tag->base[2];
	if (basemod >= 0) {
	    if (basemod == 0) {
	        im = m;
	        for (j = 0; ((j < nofrecs) & (tdadr[j] != baseadr)); j++);
		if (j > i) {
		    /* change order and try again */
		    u = tdadr[i]; tdadr[i] = tdadr[j]; tdadr[j] = u;
                    continue;
		}
	    }else im = m->imports[basemod - 1];

	    base = (Type) *(long*)(im->data + baseadr);
	    j = 0;
	    while (tag->base[j] = base->base[j]) j++;
	}
	tag->level = j; 
	tag->base[j] = (long)tag;
	while (j < maxlevel) tag->base[++j] = 0;
	for (j = 0; j < nofmth; j++) {
	    mthp = (long*)(((long)tag) - 8 - j*4);
	    if ((entno = *mthp) < 0) { /*inherited*/
		*mthp = *(long*)(((long)base) - 8 - j*4);
	    }else{ /*overwritten*/
		*mthp = (long)(m->code + m->entries[entno]);
	    }
	}
        i++;
    }
}


static void LoadScope( File f, Export scope, short level, long adr, 
		       Export eparr[], short* ix ) {
    short no, no2;
    long fp, off, arsize;
    Export dsc, elem;

    scope->nofExp = Read2(f); no = 0; no2 = 0;
    if (scope->nofExp > 0) {
        arsize = scope->nofExp * sizeof( ExportDesc );
        scope->dsc = (Export)memalign( 8, arsize );
        dsc = scope->dsc;  dsc->adr = adr;
    }
    if (level == EUrecScope) {
        (*ix)++;
        eparr[*ix] = scope; 
    }
    fp = ReadNum(f);
    while (fp != EUEnd) {
        if (fp == EURecord) {
            off = ReadNum(f);
            elem = &dsc[no2];
            if (off < 0) {
                /* old type */
                elem->nofExp = eparr[-off]->nofExp; 
                elem->dsc    = eparr[-off]->dsc;
            } else { 
                LoadScope( f, elem, EUrecScope, off, eparr, ix );
            }
        } else {
            elem = &dsc[no];
            if (level == EUobjScope) {
                elem->adr = ReadNum(f);
            }
            elem->fp = fp; no2 = no; no++;
        }
        fp = ReadNum(f);
    }
}

void LoadExpBlock( File f, Module m ) {
    Export eparr[1024]; 
    short ix;

    ix = 0; 
    m->export = (Export)memalign( 8, sizeof( ExportDesc ) );
    LoadScope( f, m->export, EUobjScope, 0, eparr, &ix );
}

static void CheckScope( File f, Module M, Module m, Export scope, 
			short level, short *res ) {
    long fp, link, i, tadr, tdadr, fadr, dadr; 
    Export elem, dsc; 
    int tmpErr;
    char name[64];

    tmpErr = (level == EUerrScope);
    dsc = scope->dsc; i = 0;
    fp = ReadNum(f);
    while (fp != EUEnd ) {
        if (fp == EURecord) { 
            link = ReadNum(f);
            elem  =  &dsc[i];
            if (tmpErr) {
                 CheckScope( f, M, m, elem, EUerrScope, res );
            } else {
                if (link >= 0) {
                    tadr = (elem->dsc->adr) & (~EUTypFlag); 
                    tdadr = *(long*)(m->data + tadr);
		    *(long*)(M->data + link) = tdadr;
                }
                CheckScope(  f, M, m, elem, EUrecScope, res );
            }
        } else { 
            ReadName( f, name );
            if (level >= EUobjScope) {
                tmpErr = false;
                if (level == EUobjScope)  link = ReadNum(f); 
                elem = dsc;
                i = 0; 
                while ((i < scope->nofExp) & (elem->fp != fp) ) {
                    i++; elem = &dsc[i];
                }
                if (i >= scope->nofExp) { 
                    if (*res == 0) {
                        *res = incompImport;
			strcpy( n1, M->name); strcpy( n2, m->name);
			strcpy( n3, name);
                    }
                    tmpErr = true;
                } else if ((level == EUobjScope) & (link != 0)) {
                    if (( link & EUProcFlag ) != EUProcFlag) {
                        *(long*)(M->data + link) = (long)(m->data - elem->adr); 
                    } else {
		        fadr = link & (~EUProcFlag);
			dadr = (elem->adr & (~EUProcFlag)) + (long)m->code;
		        FixProc( M, fadr, dadr );
                    }
                }
            }
        }
        fp = ReadNum(f);
    }
}


static void CheckUseBlock( File f, Module M, short *res ) {
    Module mod;
    char name[64];

    ReadName( f, name );
    while ((name[0] != 0) & (*res == 0)) {
        Oberon_ThisMod( name, 0, &mod, res );
        if (*res == 0) {
            CheckScope( f, M, mod, mod->export, EUobjScope, res );
        }
        ReadName( f, name );
    }
}

static void LoadModule(File f, Module *mod, short *res) { 

    Module m, im;
    long size, i, mno;
    unsigned short nofLinks;
    char *data;
    impMod mods[256];
    char imname[64];
    LinkDesc links[1000];
    char dum;
    void (*body)();
    int n;

    *mod = NULL;

    
    /* header block */
    if (Check(f, 0xBB, res) != true) /* obj flag */ {
        *res = invalidObjFile; fclose(f->handle); return;
    }
    if (Check2(f, 0xBF, 0x66, res) != true) /* SPARC obj */ {
        *res = corruptedObjFile; fclose(f->handle); return;
    }

    m = (Module) ((long)(calloc(sizeof(ModuleDesc)+8, 1)) + 8);
    if ((long) m == 4) {
        *res = notEnoughSpace; fclose(f->handle); return;
    }
    *(long*)((long) m-4) = 1; /* set mark bit */
    
    size = ReadNum(f);
    for (i=0; i < size; i++) dum = Read1(f);
    
    m->refSize = Read4(f);
    m->nofEntries = Read2(f); 
    m->nofCmds = Read2(f);
    m->nofPtrs = Read2(f); 
    m->nofTds = Read2(f); 
    m->nofImps = Read2(f);
    nofLinks = Read2(f);
    m->dataSize = Read4(f); 
    m->conSize = Read2(f);
    m->codeSize = Read2u(f); 
    ReadName( f, m->name );

    size = m->nofEntries * 4 + m->nofCmds * 36 + m->nofPtrs * 4 
           + (m->nofImps + 1) * 4 + m->dataSize + 1024 
	   + m->conSize + m->codeSize + m->refSize;
    
    size = ((size + pagesize - 1)/pagesize)*pagesize;
    m->entries = (long*)memalign(pagesize, size);
    memset( m->entries, 0, size );


    if (m->entries == NULL) {
        *res = notEnoughSpace; fclose(f->handle); return;
    }
    m->cmds = (Command)(m->entries + m->nofEntries);
    m->ptrTab = (long *)(m->cmds + m->nofCmds);
    m->imports = (Module *)(m->ptrTab + m->nofPtrs);
    data = ((char*)(m->imports + m->nofImps + 1)) + m->dataSize + 1024;
    m->data = (char*)(((long) data) & ~1023);
    m->code = data + m->conSize;
    m->refs = m->code + m->codeSize;

    m->init = false; m->refcnt = -1;

    m->next = modules; modules = m;

    /* entry block */
    if (Check(f, 0x82, res)) {
        for (i = 0; i < m->nofEntries; i++)  m->entries[i] = Read2u(f);
    }

    /* command block */
    if (Check(f, 0x83, res)) {
        for (i = 0; i < m->nofCmds; i++) {
            ReadName(f, m->cmds[i].name); m->cmds[i].entry = Read2u(f);
        }
    }

    /* pointer block */
    if (Check(f, 0x84, res)) {
        fread(m->ptrTab, 4, m->nofPtrs, f->handle); 
        f->pos = f->pos + 4*m->nofPtrs;
    }

    /* import block */
    mno = 0;
    if (Check(f, 0x85, res)) {
        while ((mno < m->nofImps) & (*res == done)) {
            ReadName(f, imname);
            Oberon_ThisMod( imname, 0, &im, res );
            if (*res == done) {
                im->refcnt++;
                m->imports[mno] = im;
                mno++;
	    }
        }
    }

    /* link block */
    if (Check(f, 0x86, res)) {
        for (i = 0; i < nofLinks; i++) {
            links[i].m = Read1(f); 
            links[i].n = Read1(f); 
            links[i].fixups = Read2u(f);
        }
    }

    /* data block */
    if (Check(f, 0x87, res)) {
        fread(m->data, 1, m->conSize, f->handle); f->pos = f->pos + m->conSize;
    }


    /* export block */
    if (Check( f, 0x88, res )) {
        LoadExpBlock( f, m );
    }

    /* code block */
    if (Check(f, 0x89, res)) {
        fread(m->code, 4, m->codeSize / 4, f->handle);f->pos = f->pos + m->codeSize;
    }

    /* use block */
    if (Check( f, 0x8A, res)) {
        CheckUseBlock( f, m, res );
    }

    if (*res == done) {

        /* type block */
        if (Check(f, 0x8B, res)) {
            ReadTypes(f, m, m->nofTds, res);
	}

        /* ref block */
        if (Check(f, 0x8D, res)) {
            fread(m->refs, 1, m->refSize, f->handle); f->pos = f->pos + m->refSize;
	}

        fclose(f->handle);
        Fixup(m, nofLinks, links); 

        if (!strcmp(m->name, "Glue")) InitGlue(m);
        m->refcnt = 0;
        m->init = true;
	
	if (memcntl( (caddr_t)m->entries, size, 
            MC_SYNC, (caddr_t)(MS_SYNC|MS_INVALIDATE),
            PRIVATE|PROT_READ|PROT_WRITE|PROT_EXEC, 0 ) < 0 ) 
            perror("oberon: mencntl");
        
	body = (void(*)())m->code;
        body();
        *res = done; 
        *mod = m;
    } else {
        while (mno > 0) {
            mno--;
            im = m->imports[mno];
            im->refcnt--;
        }
        Unlink( m );
        close(f);
        *mod = 0;
    }
}

static void GetDir(short *pos, char *dir) { 

    short i, p; char ch;

    p = *pos; ch = OBERON[p]; i = 0;
    while ((ch != ':') && (ch != 0)) {dir[i] = ch; p++; i++; ch = OBERON[p];}
    dir[i] = 0; *pos = p + 1;
}

static void FindFile(char *fname, File f, short *res) { 

    char dir[256];
    short pos, l;

    pos = 0; GetDir(&pos, dir);
    while ((dir[0] != 0) && 
        (f->handle = fopen(strcat(strcat(dir, "/"), fname), "r")) == NULL)
        GetDir(&pos, dir);
    if (f->handle == NULL) {
        *res = fileNotFound; strcpy(n1, fname);
    }
    else {
        l = strlen(dir);
	if (l > 32) {
	   strcpy( f->name, &dir[l-32] );
	   f->name[0] = '.';
	   f->name[1] = '.';
	   f->name[2] = '.';
        }
    	else strcpy( f->name, dir );
	f->pos = 0;
        *res = done;
    }
}

static Module FindModule(char *name) { 

    Module m;

    m = modules;
    while ((m != NULL) && (strcmp(m->name, name) != 0)) {
    	m = m->next;
    }
    return m;
}

static void Oberon_ThisMod(char *name, long len, Module *m, short *res) { 

    FileDesc f;
    char objFilename[64];

    *res = done; strcpy(n1, name);
    *m = FindModule(name);
    if (*m == NULL) {
        strcpy(objFilename, name);
        strcat(objFilename, ".Obj");
        FindFile(objFilename, &f, res);
        if (*res == done) LoadModule(&f, m, res);
    }
}

static void Oberon_ThisCommand( Module m, char *cmdname, long length, 
				long *adr, short *res) { 
    short i;

    i = 0;
    while (i < m->nofCmds && strcmp(m->cmds[i].name, cmdname) != 0) i++;
    if (i == m->nofCmds) {
        *adr = 0; *res = cmdNotFound;
        strcpy(n2, m->name);
        strcpy(n3, cmdname);
    } else {
        *adr = (long)m->code + m->cmds[i].entry;
        *res = done;
    }
}


static void Init() { 
    void *libc;

    libc = Oberon_dlopen("libc.so.1", 0, 999);
    Oberon_dlsym(libc, ".mul",    (void*)&mul);
    Oberon_dlsym(libc, "__mul64", (void*)&mulH);
    RtoH  = (long)__ftoll;
    LRtoH = (long)__dtoll;
    HtoR  = (long)__floatdisf;
    HtoLR = (long)__floatdidf;
    modules = NULL;
    OBERON = getenv("OBERON");
    if (OBERON == NULL) OBERON = ".:/usr/local/Oberon";
}



static void FreeModule(Module m, char all) { 

    int i; 

    for (i = 0; (i < m->nofImps) && (m->imports[i] != NULL); i++) {
        m->imports[i]->refcnt--;
        if ((m->imports[i]->refcnt == 0) && all) 
            FreeModule(m->imports[i], all);
    }
    Unlink(m);
}

static void Oberon_Free(char *name, long length, char all, short *res) {

    Module m;

    *res = done; strcpy(n1, name);
    m = FindModule(name); 
    if (m != NULL) {
        if (m->refcnt > 0) *res = refCntNotZero;
        else FreeModule(m, all);
    } else *res = moduleNotFound;
}




main(long argc, char **argv) { 

    void (*proc)();
    Module m;
    short res;
    char msg[128];
    ac = argc;
    av = argv;

    /* check if we have suid root previlliges */
    if (geteuid() == 0) {
        suid_root = true;
        seteuid( getuid() );
    } else {
        suid_root = false;
    }
    pagesize = sysconf(_SC_PAGESIZE);

    printf("Oberon Loader 20.10.2003\n");

    printf("pagesize = %d\n", pagesize);

    Init();

    /* load base modules up to Kernel */
    Oberon_ThisMod("Glue", 0, &m, &res); 
    if (res == done) Oberon_ThisMod("Console", 0, &m, &res);
    if (res == done) Oberon_ThisMod("Unix", 0, &m, &res);
    if (res == done) Oberon_ThisMod("Threads", 0, &m, &res);
    if (res == done) Oberon_ThisMod("Kernel", 0, &m, &res);
    if (res == done) Oberon_ThisMod("Modules", 0, &m, &res);
    /*
    body of 'Modules' starts Oberon.Loop
    or a command specified by the option:
    		-x Module.Procedure
    */
    printf("loading Oberon failed\n");
    exit(1);
}
