/*------------------------------------------------------
 * Oberon Boot File Loader RC, JS 27.4.93/2.12.93, HP-UX 9.0 Version
 *
 * Oberon Boot File Loader for Linux
 * derived from HP and Windows Boot Loader
 * MAD, 23.05.94
 * PR,  01.02.95  support for sockets added
 * PR,  05.02.95  support for V24 added
 * PR,  23.12.95  migration to shared ELF libraries
 * RLI, 22.08.96  added some math primitives
 * RLI, 27.01.97  included pixmap
 * RLI, 13.10.97  changed name of Fontmap - File
 * g.f. 01.11.99  added InstallTraphandler
 *		  added Threads support
 *		  removed cmd line parameter evaluation
 * g.f. 22.11.04  call to mprotect added
 *
 * without Threads support:
 * Linux:	gcc -o linux.oberon   -DLINUX   oberon.c -ldl -lm
 * Solaris:	gcc -o solaris.oberon -DSOLARIS oberon.c -ldl -lrt -lm
 *
 * with Threads support:
 * Linux:	gcc -o linux.oberon   -DLINUX   -DTHREADS oberon.c -ldl -lpthread -lm
 * Solaris:	gcc -o solaris.oberon -DSOLARIS -DTHREADS oberon.c -ldl -lpthread -lrt -lm
 *
 *-----------------------------------------------------------*/

#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <math.h>		/* RLI */
#include <setjmp.h>		/* g.f. */
#include <signal.h>		/* g.f. */
#include <limits.h>		/* g.f. */
#include "Threads.h"		/* g.f. */
#include <sys/mman.h>


typedef unsigned int	uint;
typedef unsigned int	addr;

typedef void (*Proc)();

FILE *fd;
char *OBERON;
char path[4096];
char *dirs[255];
char fullname[512];
int nofdir;
char defaultpath[] = ".:/usr/local/Oberon:/usr/local/Oberon/.Fonts";
#ifdef SOLARIS
char bootname[64] = "SolarisOberonCore";
#else
char bootname[64] = "LinuxOberonCore";
#endif
uint heapSize;
addr heapAdr;
int Argc;
char **Argv;
int debug;
int suid_root;

#define BLSIZE	4096

typedef	void(*traphandler_t)(void*, void*, int);

static traphandler_t	OberonTrapHandler;

static void traphandler(int sig, void *scp, void *ucp) {
	
	/*printf("traphandler %d %x %x\n", sig, scp, ucp);*/
	/*OberonTrapHandler(sig, scp, ucp);*/
	OberonTrapHandler(ucp, scp, sig); /* rev. order: Oberon <--> C */

	/*printf("returned fron Oberon Trap Handler\n");*/
	return;
}

static void installHandler(int sig) {
	struct sigaction act;
	sigset_t mask;

	sigemptyset(&mask);
	act.sa_mask = mask;
#ifdef SOLARIS
	act.sa_flags =  SA_SIGINFO|SA_ONSTACK|SA_NODEFER;
	act.sa_handler = traphandler;
#else
	act.sa_flags =  SA_SIGINFO|SA_NODEFER;
	act.sa_handler = (__sighandler_t)traphandler;
#endif
	if (sigaction( sig, &act, NULL ) != 0) {
		perror("sigaction");
	}
}

static void InstallTrapHandler(traphandler_t p) {
	int i;
	
	/*printf("traphandler = %x\n", p);*/
	OberonTrapHandler = p;

	for (i = 1; i <= 15; i++) {
	     if (i != 9) installHandler( i );
	}
}

/*--------------------------------------------------- g.f. -----*/


int dl_open(char *lib, int mode)
{
  void *handle;

  if (debug) printf("dl_open: %s\n", lib);
  if ((handle = dlopen(lib, mode)) == NULL) {
    if (debug)
      printf("dl_open: %s not loaded, error = %s\n", lib, dlerror());
  }

  return (int)handle;
}

void dl_close(int handle)	/* not necessary */
{
  dlclose((void *)handle);
}

static int errnof()
{
	return errno;
}

static void o_printf(char *fs, int p1, int p2, int p3, int p4) {
	printf(fs, p1, p2, p3, p4);
	fflush(stdout);
}

int o_stat(char* name, void* buf) {
	return stat(name, (struct stat *) buf);
}

int o_lstat(char* name, void* buf) {
	return lstat(name, (struct stat *) buf);
}

int o_fstat(int fd, void* buf) {
	return fstat(fd, (struct stat *) buf);
}

int o_open(char* name, int flags, int mode) {
	int r;
	
	r = open(name, flags, mode);
	if (r < 0) { fprintf(stderr, "<%s> ", name ); perror("open");  }
	return r;
}

void dl_sym(int handle, char *symbol, int *adr)
{
  if (debug) printf("dl_sym: %s\n", symbol);
  
  if      (strcmp("dlopen",	symbol) == 0) *adr = (int)dl_open;
  else if (strcmp("dlclose",	symbol) == 0) *adr = (int)dl_close;
  else if (strcmp("debug",	symbol) == 0) *adr = debug;
  else if (strcmp("heapAdr",	symbol) == 0) *adr = heapAdr;
  else if (strcmp("heapSize",	symbol) == 0) *adr = heapSize;
  else if (strcmp("argc",	symbol) == 0) *adr = Argc;
  else if (strcmp("argv",	symbol) == 0) *adr = (int)Argv;
  else if (strcmp("exit",	symbol) == 0) *adr = (int)exit;
  else if (strcmp("printf",	symbol) == 0) *adr = (int)o_printf;

  else if (strcmp("InstallTrapHandler",
  				symbol) == 0) *adr = (int)InstallTrapHandler;
  
  /* the following synbols are missing in the dynamic version of libc
   * on Linux systems or have different names
   */
#ifdef LINUX
  else if (strcmp("sigsetjmp",	symbol) == 0) *adr = (int)__sigsetjmp;
  else if (strcmp("setjmp",	symbol) == 0) *adr = (int)__sigsetjmp;
  else if (strcmp("mknod",	symbol) == 0) *adr = (int)mknod;
  else if (strcmp("stat",	symbol) == 0) *adr = (int)o_stat;
  else if (strcmp("lstat",	symbol) == 0) *adr = (int)o_lstat;
  else if (strcmp("fstat",	symbol) == 0) *adr = (int)o_fstat;
#endif
#ifdef SOLARIS
  else if (strcmp("stat",	symbol) == 0) *adr = (int)o_stat;
  else if (strcmp("lstat",	symbol) == 0) *adr = (int)o_lstat;
  else if (strcmp("fstat",	symbol) == 0) *adr = (int)o_fstat;
#endif
  else if (strcmp("errno",	symbol) == 0) *adr = (int)errnof;
  
  /* Math.Mod stuff -- added by RLI */  
  else if (strcmp("sin",	symbol) == 0) *adr = (int)sin;
  else if (strcmp("cos",	symbol) == 0) *adr = (int)cos;
  else if (strcmp("log",	symbol) == 0) *adr = (int)log;
  else if (strcmp("atan",	symbol) == 0) *adr = (int)atan;
  else if (strcmp("exp",	symbol) == 0) *adr = (int)exp;
  else if (strcmp("sqrt",	symbol) == 0) *adr = (int)sqrt;

  /* threads support */
  else if (strcmp("mtxInit",   		symbol) == 0) *adr = (int)_mtx_init;
  else if (strcmp("mtxDestroy", 	symbol) == 0) *adr = (int)_mtx_destroy;
  else if (strcmp("mtxLock",    	symbol) == 0) *adr = (int)_mtx_lock;
  else if (strcmp("mtxUnlock",  	symbol) == 0) *adr = (int)_mtx_unlock;
  else if (strcmp("thrStart",		symbol) == 0) *adr = (int)_thr_start;
  else if (strcmp("thrThis",		symbol) == 0) *adr = (int)_thr_this;
  else if (strcmp("thrSleep",		symbol) == 0) *adr = (int)_thr_sleep;
  else if (strcmp("thrPass",		symbol) == 0) *adr = (int)_thr_pass;
  else if (strcmp("thrExit",		symbol) == 0) *adr = (int)_thr_exit;
  else if (strcmp("thrSuspend",		symbol) == 0) *adr = (int)_thr_suspend;
  else if (strcmp("thrResume",		symbol) == 0) *adr = (int)_thr_resume;
  else if (strcmp("thrGetPriority",	symbol) == 0) *adr = (int)_thr_getprio;
  else if (strcmp("thrSetPriority",	symbol) == 0) *adr = (int)_thr_setprio;
  else if (strcmp("thrMaxPriority",	symbol) == 0) *adr = (int)_thr_maxprio;
  else if (strcmp("thrOldPriority",	symbol) == 0) *adr = (int)_thr_oldprio;
  else if (strcmp("thrKill",		symbol) == 0) *adr = (int)_thr_kill;
  else if (strcmp("thrInitialize",	symbol) == 0) *adr = (int)_thr_initialize;
  else {
    *adr = (int)dlsym((void *) handle, symbol);
    if (*adr == 0) {
      printf("dl_sym: symbol %s not found\n", symbol); 
    }
  }
}


/*----- Files Reading primitives -----*/

int Rint() 

{
  unsigned char b[4];
  /*
     b[3] = fgetc(fd); b[2] = fgetc(fd); b[1] = fgetc(fd); b[0] = fgetc(fd);
     */
  /* little endian machine reading little endian integer */
  b[0] = fgetc(fd); b[1] = fgetc(fd); b[2] = fgetc(fd); b[3] = fgetc(fd);
  return *((int *) b);
}

int RNum()
{
  int n, shift;
  unsigned char x;
  shift = 0; n = 0; x = fgetc(fd);
  while (x >= 128) {
    n += (x & 0x7f) << shift;
    shift += 7;
    x = fgetc(fd);
  }
  return n + (((x & 0x3f) - ((x >> 6) << 6)) << shift);
}
	
void Relocate(uint heapAdr, int shift)
{
  int len; addr adr; 
  
  len = RNum(); 
  while (len != 0) { 
    adr = RNum(); 
    adr += heapAdr; 
    *((int *)adr) += shift; 
    len--; 
  } 
}

void showProc(addr adr) {
  int i;
  printf("Oberon code to be called:\n");
  for (i=0; i<10; i++) {
    printf("%8x: %8x\n", adr, *(int*)adr);
    adr += 4;
  }
  printf("%8x: ...\n", adr);
}

void Boot()
{
  addr adr, fileHeapAdr, dlsymAdr;
  uint len, d, codeSize, fileHeapSize;
  int shift, notfound;  
  Proc body;

  d = 0; notfound = 1;
  while ((d < nofdir) && notfound) {
    strcat(strcat(strcpy(fullname, dirs[d++]), "/"), bootname);
    fd = fopen(fullname, "r");
    if (fd != NULL) notfound = 0;
  }
  if (notfound) {
    printf("oberon: boot file %s not found\n", bootname);  
    exit(-1);
  }
  fileHeapAdr = Rint(); fileHeapSize = Rint();
  if (fileHeapSize >= heapSize) {
    printf("oberon loader: heap too small\n");  
    exit(-1);
  }
  adr = heapAdr; len = fileHeapSize + 32; 
  while (len > 0) { 
    *((int*)adr) = 0; 
    len -= 4; adr += 4; 
  } 
  shift = heapAdr - fileHeapAdr;
  adr = Rint(); len = Rint();
  while (len != 0) {
    adr += shift;
    len += adr;
    codeSize = len - heapAdr;
    while (adr != len) { *((int*)adr) = Rint(); adr += 4; }
    adr = Rint(); len = Rint();
  }
  body = (Proc)(adr + shift);
  Relocate(heapAdr, shift);
  dlsymAdr = Rint();
  if (debug) showProc((addr)body);
  *((int *)(heapAdr + dlsymAdr)) = (int)dl_sym;
  fclose(fd);
  codeSize = (codeSize+4095)/4096*4096;
  if(mprotect((void*)heapAdr, codeSize, PROT_READ|PROT_WRITE|PROT_EXEC) != 0)
     perror("mprotect");
  (*body)();
}

void InitPath()
{
  int pos;
  char ch;
  
  if ((OBERON = getenv("OBERON")) == NULL) OBERON = defaultpath;
  strcpy(path, OBERON);
  pos = 0; nofdir = 0;
  ch = path[pos++];
  while (ch != '\0') {
    while ((ch == ' ') || (ch == ':')) ch = path[pos++];
    dirs[nofdir] = &path[pos-1];
    while ((ch > ' ') && (ch != ':')) ch = path[pos++];
    path[pos-1] = '\0';
    nofdir ++;
  }
}

void doexit(int ret, void *arg)
{
  _exit(ret);
}

int main(int argc, char *argv[])
{
  char* p;
  
/*  on_exit(doexit, NULL);	*/
  
  Argc = argc; Argv = argv;

  printf("Oberon Boot File Loader 22.11.2004\n");
  /* check if we have suid root previlliges */
  if (geteuid() == 0) {
     suid_root = 1;
     seteuid( getuid() );
  } else {
     suid_root = 0;
  }

  debug = 0;
  p = getenv("OBERON_DEBUG");
  if (p != NULL) debug = atoi(p);

  heapSize = 0x200000;
  heapAdr = (addr)memalign(4096, heapSize);
  if (heapAdr == 0) {
    printf("oberon: cannot allocate initial heap space\n");  
    exit(-1);
  }
  
  InitPath();
  
  Boot();
  return 0;
}



